import numpy as np
import pandas as pd
from collections import Counter

import h5py as h5
import re
import numpy as np
import pandas as pd
import tqdm

def meta(file, search_term, meta_fields=["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"], remove_sc=False, silent=False):
    """
    Search for samples in a file based on a search term in specified metadata fields.

    Args:
        file (str): The file path or object containing the data.
        search_term (str): The term to search for. Case-insensitive.
        meta_fields (list, optional): The list of metadata fields to search within.
            Defaults to ["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"].
        remove_sc (bool, optional): Whether to remove single-cell samples from the results.
            Defaults to False.
        silent (bool, optional): Print progress bar.

    Returns:
        pd.DataFrame: DataFrame containing the extracted metadata, with metadata fields as columns and samples as rows.
    """
    search_term = search_term.upper()
    with h5.File(file, "r") as f:
        meta = []
        idx = []
        mfields = []
        for field in tqdm.tqdm(meta_fields, disable=not silent):
            if field in f["meta"]["samples"].keys():
                try:
                    meta.append([x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"][field]))])
                    mfields.append(field)
                except Exception:
                    x=0
        meta = pd.DataFrame(meta, index=mfields ,columns=[x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"]["geo_accession"]))])
        for i in tqdm.tqdm(range(meta.shape[0]), disable=silent):
            idx.extend([i for i, item in enumerate(meta.iloc[i,:]) if re.search(search_term, item.upper())])
        if remove_sc:
            singleprob = np.where(np.array(f["meta/samples/singlecellprobability"]) < 0.5)[0]
            idx = sorted(list(set(idx).intersection(set(singleprob))))
    return meta.iloc[:,idx].T

def samples(file, samples, meta_fields=["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"], silent=False):
    """
    Extracts metadata for specified samples from an HDF5 file.

    Args:
        file (str): Path to the HDF5 file.
        samples (list): List of samples to extract metadata for.
        meta_fields (list, optional): List of metadata fields to extract. Defaults to ["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"].
        silent (bool, optional): If True, disables the progress bar. Defaults to False.

    Returns:
        pandas.DataFrame: DataFrame containing the extracted metadata, with metadata fields as columns and samples as rows.
    """
    samples = set(samples)
    with h5.File(file, "r") as f:
        meta = []
        mfields = []
        meta_samples = np.array([x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"]["geo_accession"]))])
        idx = [i for i,x in enumerate(meta_samples) if x in samples]
        for field in tqdm.tqdm(meta_fields, disable=not silent):
            if field in f["meta"]["samples"].keys():
                try:
                    meta.append([x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"][field][idx]))])
                    mfields.append(field)
                except Exception:
                    x=0
        meta = pd.DataFrame(meta, index=mfields ,columns=[x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"]["geo_accession"][idx]))])
        inter = meta.columns.intersection(set(samples))
    return meta.loc[:,inter].T

def series(file, series, meta_fields=["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"], silent=False):
    """
    Extracts metadata for specified series from an HDF5 file.

    Args:
        file (str): Path to the HDF5 file.
        series: Series to extract metadata for.
        meta_fields (list, optional): List of metadata fields to extract. Defaults to ["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"].
        silent (bool, optional): If True, disables the progress bar. Defaults to False.

    Returns:
        pandas.DataFrame: DataFrame containing the extracted metadata, with metadata fields as columns and samples as rows.
    """
    with h5.File(file, "r") as f:
        meta = []
        mfields = []
        meta_series = np.array([x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"]["series_id"]))])
        idx = [i for i,x in enumerate(meta_series) if x == series]
        for field in tqdm.tqdm(meta_fields, disable=not silent):
            if field in f["meta"]["samples"].keys():
                try:
                    meta.append([x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"][field][idx]))])
                    mfields.append(field)
                except Exception:
                    x=0
        meta = pd.DataFrame(meta, index=mfields ,columns=[x.decode("UTF-8").upper() for x in list(np.array(f["meta"]["samples"]["geo_accession"][idx]))])
    return meta.T

def get_meta(file):
    f = h5.File(file, "r")
    meta = {}
    for field in list(f["meta"]["samples"].keys()):
        meta[field] = [x.decode("UTF-8") for x in list(np.array(f["meta"]["samples"][field]))]
    f.close()
    return meta

def get_meta_sample_field(file, field):
    f = h5.File(file, "r")
    meta_data = [x.decode("UTF-8") for x in list(np.array(f["meta"]["samples"][field]))]
    f.close()
    return meta_data

def get_meta_gene_field(file, field):
    f = h5.File(file, "r")
    meta_data = [x.decode("UTF-8") for x in list(np.array(f["meta"]["genes"][field]))]
    f.close()
    return meta_data



import numpy as np
import pandas as pd
from scipy import stats

import os
import json

import qnorm

def get_config():
    config_url = os.path.join(
        os.path.dirname(__file__),
        'data/config.json')
    with open(config_url) as json_file:
        data = json.load(json_file)
    return(data)


def versions():
    """
    Get the available versions of human gene counts.

    Returns:
        list: A list of available versions of human gene counts.
    """
    config = get_config()
    versions = config["GENE_COUNTS"]["HUMAN"].keys()
    return versions

def normalize(counts, method="log_quantile", tmm_outlier=0.05):
    """
    Normalize the count matrix using a specified method.

    Args:
        counts (pd.DataFrame): A pandas DataFrame representing the count matrix.
        method (str, optional): The normalization method to be applied. Default is "log_quantiles".
            - "quantile": Perform quantile normalization on the counts.
            - "log_quantile": Perform quantile normalization on the log-transformed counts.
            - "cpm": Perform count per million (CPM) normalization.
            - "tmm": Perform trimmed mean normalization

    Returns:
        pd.DataFrame: A normalized count matrix as a pandas DataFrame with the same index and columns as the input.

    Raises:
        ValueError: If an unsupported normalization method is provided.
    """
    norm_exp = 0
    if method == "quantile":
        norm_exp = qnorm.quantile_normalize(np.array(counts))
    elif method == "log_quantile":
        norm_exp = qnorm.quantile_normalize(np.log2(1+np.array(counts)))
    elif method == "cpm":
        g = np.array(counts)
        norm_exp = np.abs(g/g.sum(axis=0))*1_000_000
    elif method == "tmm":
        norm_exp = tmm_norm(counts, tmm_outlier)
    else:
        raise ValueError("Unsupported normalization method: " + method)
    norm_exp = pd.DataFrame(norm_exp, index=counts.index, columns=counts.columns, dtype=np.float32)
    return norm_exp

def tmm_norm(exp, percentage=0.05):
    lexp = np.log2(1+exp).astype(np.float32)
    tmm = trimmed_mean(lexp, percentage)
    nf = pd.DataFrame(np.tile(tmm, (exp.shape[0], 1)), index=lexp.index, columns=lexp.columns)
    temp = (lexp/nf)
    return temp

def trimmed_mean(matrix, percentage):
    matrix = np.array(matrix)
    trimmed_means = []
    for col in range(matrix.shape[1]):
        data = matrix[:, col].copy()
        data = data[data > 0]
        n_trim = int(len(data) * percentage)
        sorted_values = np.sort(data)
        trimmed_values = sorted_values[n_trim:-n_trim]
        trimmed_mean = np.mean(trimmed_values)
        trimmed_means.append(trimmed_mean)
    return trimmed_means


import wget
import sys
import requests
from tqdm import tqdm
import archs4py.utils

def bar_progress(current, total, width=80, update_interval=10):
    current_gb = current / (1024**3)  # Convert current bytes to GB
    total_gb = total / (1024**3)  # Convert total bytes to GB
    
    if current % (update_interval * 1024**2) == 0:  # Update progress every 10 MB
        progress_message = "Downloading: %d%% [%.2f GB / %.2f GB]" % (current / total * 100, current_gb, total_gb)
        sys.stdout.write("\r" + progress_message)
        sys.stdout.flush()

def counts(species, path="", type="GENE_COUNTS", version="latest"):
    """
    Download count files for a given species and count type.

    Args:
        species (str): The species for which count files are being downloaded. ["human", "mouse"]
        path (str, optional): The path where the downloaded file will be saved. Defaults to "".
        type (str, optional): The type of count file to be downloaded. Defaults to "GENE_COUNTS".
        version (str, optional): The version of the count file to be downloaded. Defaults to "latest". Versions can be listed with archs4py.versions()

    Returns:
        str: The path where the count file is downloaded.

    Raises:
        Exception: If an error occurs during the download process.

    Notes:
        The function first tries to download the count file using the primary URL specified in the configuration file.
        If the download fails, it falls back to the fallback URL specified in the configuration file.

        Supported count types:
        - GENE_COUNTS: Gene-level count files.
        - TRANSCRIPT_COUNTS: Transcript-level count files.
    """
    conf = archs4py.utils.get_config()

    try:
        fpath = wget.download(conf[type][species.upper()][version]["primary"], out=path, bar=bar_progress)
        print("file downloaded to", fpath)
    except Exception:
        fpath = wget.download(conf[type][species.upper()][version]["fallback"], out=path, bar=bar_progress)
        print("file downloaded to", fpath)




import archs4py.data
import archs4py.download
import archs4py.meta
import archs4py.utils

import importlib
importlib.reload(archs4py.data)
importlib.reload(archs4py.download)
importlib.reload(archs4py.meta)
importlib.reload(archs4py.utils)

from archs4py.utils import versions
from archs4py.utils import normalize





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="archs4py",
    version="0.1.11",
    author="Alexander Lachmann",
    author_email="alexander.lachmann@mssm.edu",
    description="ARCHS4 python package supporting data loading and data queries.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/maayanlab/archs4py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_data={
        "archs4py": ["data/*"]
    },
    include_package_data=True,
    install_requires=[
        'h5py==3.8.0',
        'numpy==1.24.3',
        'pandas==2.0.2',
        'qnorm==0.8.1',
        'setuptools==67.8.0',
        'tqdm==4.65.0',
        'wget==3.2',
        's3fs==2023.5.0',
    ],
    python_requires='>=3.7',
)


import numpy as np
import pandas as pd

import h5py as h5
import s3fs
import tqdm
import re

import multiprocessing
import random

def resolve_url(url):
    u1 = url.rsplit('/', 1)
    u2 = u1[0].rsplit('/', 1)
    file_name = u1[-1]
    bucket_name = u2[-1]
    endpoint = u2[0]
    S3_URL = "s3://"+bucket_name+"/"+file_name
    return(S3_URL, endpoint)

def fetch_meta_remote(field, s3_url, endpoint):
    s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        meta = [x.decode("UTF-8") for x in list(np.array(f[field]))]
    return np.array(meta)

def meta(file, search_term, meta_fields=["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"],  remove_sc=False, silent=False):
    """
    Search for samples in a file based on a search term in specified metadata fields.

    Args:
        file (str): The file path or object containing the data.
        search_term (str): The term to search for. The search is case-insensitive and supports regular expressions.
        meta_fields (list, optional): The list of metadata fields to search within.
            Defaults to ["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"].
        remove_sc (bool, optional): Whether to filter single-cell samples from the results.
            Defaults to False.
        silent (bool, optional): Print progress bar.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the gene expression data for the matching samples.
    """
    search_term = re.sub(r"_|-|'|/| |\.", "", search_term.upper())
    if not silent:
        print("Searches for any occurrence of", search_term, "as regular expression")
    if file.startswith("http"):
        return meta_remote(file, search_term, meta_fields, remove_sc, silent)
    else:
        return meta_local(file, search_term, meta_fields, remove_sc, silent)

def meta_local(file, search_term, meta_fields=["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"], remove_sc=False, silent=False):
    f = h5.File(file, "r")
    idx = []
    for field in meta_fields:
        if field in f["meta"]["samples"].keys():
            meta = [x.decode("UTF-8") for x in list(np.array(f["meta"]["samples"][field]))]
            idx.extend([i for i, item in enumerate(meta) if re.search(search_term, re.sub(r"_|-|'|/| |\.", "", item.upper()))])
    if remove_sc:
        singleprob = np.where(np.array(f["meta/samples/singlecellprobability"]) < 0.5)[0]
    f.close()
    if remove_sc:
        idx = sorted(list(set(idx).intersection(set(singleprob))))
    else:
        idx = sorted(list(set(idx)))
    counts = index(file, idx, silent=silent)
    return counts

def meta_remote(url, search_term, meta_fields=["geo_accession", "series_id", "characteristics_ch1", "extract_protocol_ch1", "source_name_ch1", "title"], remove_sc=False, silent=False):
    s3_url, endpoint = resolve_url(url)
    idx = []
    s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        for field in meta_fields:
            if field in f["meta"]["samples"].keys():
                meta = [x.decode("UTF-8") for x in list(np.array(f["meta"]["samples"][field]))]
                idx.extend([i for i, item in enumerate(meta) if re.search(search_term, re.sub(r"_|-|'|/| |\.", "", item.upper()))])
        if remove_sc:
            singleprob = np.where(np.array(f["meta/samples/singlecellprobability"]) < 0.5)[0]
    if remove_sc:
        idx = sorted(list(set(idx).intersection(set(singleprob))))
    else:
        idx = sorted(list(set(idx)))
    counts = index_remote(url, idx, silent)
    return counts

def rand(file, number, seed=1, remove_sc=False, silent=False):
    """
    Randomly select a specified number of samples from a file.

    Args:
        file (str): The file path or object containing the data.
        number (int): The number of samples to select randomly.
        seed (int, optional): The seed value for the random number generator. Defaults to 1.
        remove_sc (bool, optional): Whether to remove single-cell samples from the selection. Defaults to False.
        silent (bool, optional): Print progress bar.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the randomly selected samples' gene expression data.
    """
    random.seed(seed)
    if file.startswith("http"):
        return rand_remote(file, number, remove_sc, silent)
    else:
        return rand_local(file, number, remove_sc, silent)

def rand_local(file, number, remove_sc, silent=False):
    f = h5.File(file, "r")
    gsm_ids = [x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])]
    if remove_sc:
        singleprob = np.array(f["meta/samples/singlecellprobability"])
    f.close()
    if remove_sc:
        idx = sorted(random.sample(list(np.where(singleprob < 0.5)[0]), number))
    else:
        idx = sorted(random.sample(range(len(gsm_ids)), number))
    return index(file, idx, silent=silent)

def rand_remote(url, number, remove_sc, silent=False):
    s3_url, endpoint = resolve_url(url)
    s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        number_samples = len(f["meta/samples/geo_accession"])
        if remove_sc:
            singleprob = np.array(f["meta/samples/singlecellprobability"])
    if remove_sc:
        idx = sorted(random.sample(list(np.where(singleprob < 0.5)[0]), number))
    else:
        idx = sorted(random.sample(range(number_samples), number))
    return index_remote(url, idx, silent=silent)

def series(file, series_id, silent=False):
    """
    Retrieve samples belonging to a specific series from a file.

    Args:
        file (str): The file path or object containing the data.
        series_id (str): The ID of the series to retrieve samples from.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the gene expression data for the samples belonging to the specified series.
    """
    if file.startswith("http"):
        return series_remote(file, series_id, silent=silent)
    else:
        return series_local(file, series_id, silent=silent)

def series_local(file, series_id, silent=False):
    f = h5.File(file, "r")
    series = [x.decode("UTF-8") for x in np.array(f["meta/samples/series_id"])]
    f.close()
    idx = [i for i,x in enumerate(series) if x == series_id]
    if len(idx) > 0:
        return index(file, idx, silent=silent)

def series_remote(url, series_id, silent=False):
    s3_url, endpoint = resolve_url(url)
    s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        series = [x.decode("UTF-8") for x in np.array(f["meta/samples/series_id"])]
    idx = [i for i,x in enumerate(series) if x == series_id]
    if len(idx) > 0:
        return index_remote(url, idx, silent)

def samples(file, sample_ids, silent=False):
    if file.startswith("http"):
        return samples_remote(file, sample_ids, silent=silent)
    else:
        return samples_local(file, sample_ids, silent=silent)

def samples_local(file, sample_ids, silent=False):
    sample_ids = set(sample_ids)
    f = h5.File(file, "r")
    samples = [x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])]
    f.close()
    idx = [i for i,x in enumerate(samples) if x in sample_ids]
    if len(idx) > 0:
        return index(file, idx, silent=silent)

def samples_remote(url, sample_ids, silent=False):
    sample_ids = set(sample_ids)
    s3_url, endpoint = resolve_url(url)
    s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        sample_ids = [x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])]
    idx = [i for i,x in enumerate(samples) if x in sample_ids]
    if len(idx) > 0:
        return index_remote(url, idx, silent=silent)

def index(file, sample_idx, gene_idx = [], silent=False):
    """
    Retrieve gene expression data from a specified file for the given sample and gene indices.

    Args:
        file (str): The file path or object containing the data.
        sample_idx (list): A list of sample indices to retrieve expression data for.
        gene_idx (list, optional): A list of gene indices to retrieve expression data for. Defaults to an empty list (return all).
        silent (bool, optional): Whether to disable progress bar. Defaults to False.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the gene expression data.
    """
    sample_idx = sorted(sample_idx)
    gene_idx = sorted(gene_idx)
    row_encoding = get_encoding(file)
    f = h5.File(file, "r")
    genes = np.array([x.decode("UTF-8") for x in np.array(f[row_encoding])])
    if len(sample_idx) == 0:
        return pd.DataFrame(index=genes[gene_idx])
    gsm_ids = np.array([x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])])[sample_idx]
    f.close()
    if len(gene_idx) == 0:
        gene_idx = list(range(len(genes)))
    exp = []
    PROCESSES = 16
    with multiprocessing.Pool(PROCESSES) as pool:
        results = [pool.apply_async(get_sample, (file, i, gene_idx)) for i in sample_idx]
        for r in tqdm.tqdm(results, disable=silent):
            res = r.get()
            exp.append(res)
    exp = np.array(exp).T
    exp = pd.DataFrame(exp, index=genes[gene_idx], columns=gsm_ids, dtype=np.uint32)
    return exp

def index_remote(url, sample_idx, gene_idx = [], silent=False):
    if len(sample_idx) == 0:
        return pd.DataFrame(index=genes[gene_idx])
    s3_url, endpoint = resolve_url(url)
    sample_idx = sorted(sample_idx)
    gene_idx = sorted(gene_idx)
    s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
    row_encoding = get_encoding_remote(s3, url)
    genes = fetch_meta_remote(row_encoding, s3_url, endpoint)
    if len(gene_idx) == 0:
        gene_idx = np.array(list(range(len(genes))))
    gsm_ids = fetch_meta_remote("meta/samples/geo_accession", s3_url, endpoint)[sample_idx]
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        exp = np.array(f["data/expression"][:,np.array(sample_idx)], dtype=np.uint32)[gene_idx]
    exp = pd.DataFrame(exp, index=genes[gene_idx], columns=gsm_ids, dtype=np.uint32)
    return exp

def get_sample(file, i, gene_idx):
    try:
        f = h5.File(file, "r")
        temp = np.array(f["data/expression"][:,i], dtype=np.uint32)[gene_idx]
        f.close()
    except Exception:
        dd = np.array([0]*len(gene_idx))
        return dd
    return temp

def get_sample_remote(s3_url, endpoint, i, gene_idx):
    try:
        s3 = s3fs.S3FileSystem(anon=True, client_kwargs={'endpoint_url': endpoint})
        with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
            temp = np.array(f["data/expression"][:,i], dtype=np.uint32)[gene_idx]
        return temp
    except Exception:
        dd = np.array([0]*len(gene_idx))
        return dd

def get_encoding(file):
    with h5.File(file) as f:
        if "genes" in list(f["meta"].keys()):
            if "gene_symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/gene_symbol"
            elif "symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/symbol"
        elif "transcripts" in list(f["meta"].keys()):
            if "ensembl_id" in list(f["meta/trancripts"].keys()):
                return "meta/trancripts/ensembl_id"
        else:
            raise Exception("error in gene/transcript meta data")

def get_encoding_remote(s3, s3_url):
    with h5.File(s3.open(s3_url, 'rb'), 'r', lib_version='latest') as f:
        if "genes" in list(f["meta"].keys()):
            if "gene_symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/gene_symbol"
            elif "symbol" in list(f["meta/genes"].keys()):
                return "meta/gene_symbol"
        elif "transcripts" in list(f["meta"].keys()):
            if "ensembl_id" in list(f["meta/trancripts"].keys()):
                return "meta/trancripts/ensembl_id"
        else:
            raise Exception("error in gene/transcript meta data")